/*
 * This file is part of the Score-P software (http://www.score-p.org)
 *
 * Copyright (c) 2009-2012,
 * RWTH Aachen University, Germany
 *
 * Copyright (c) 2009-2012,
 * Gesellschaft fuer numerische Simulation mbH Braunschweig, Germany
 *
 * Copyright (c) 2009-2012, 2014,
 * Technische Universitaet Dresden, Germany
 *
 * Copyright (c) 2009-2012,
 * University of Oregon, Eugene, USA
 *
 * Copyright (c) 2009-2012,
 * Forschungszentrum Juelich GmbH, Germany
 *
 * Copyright (c) 2009-2012,
 * German Research School for Simulation Sciences GmbH, Juelich/Aachen, Germany
 *
 * Copyright (c) 2009-2012,
 * Technische Universitaet Muenchen, Germany
 *
 * This software may be modified and distributed under the terms of
 * a BSD-style license.  See the COPYING file in the package base
 * directory for details.
 *
 */


/**
 *  @file
 *  @source     templates/otf2_attribute_list_inc.tmpl.c
 *
 *  @brief      Convenient helper function to add and get attributes.
 */


OTF2_ErrorCode
OTF2_AttributeList_AddUint8( OTF2_AttributeList* attributeList,
                             OTF2_AttributeRef   attributeID,
                             uint8_t             uint8Value )
{
    OTF2_AttributeValue attribute_value;
    attribute_value.uint8 = uint8Value;

    /* Let OTF2_AttributeList_AddAttribute do the error checking */
    return OTF2_AttributeList_AddAttribute( attributeList,
                                            attributeID,
                                            OTF2_TYPE_UINT8,
                                            attribute_value );
}


OTF2_ErrorCode
OTF2_AttributeList_AddUint16( OTF2_AttributeList* attributeList,
                              OTF2_AttributeRef   attributeID,
                              uint16_t            uint16Value )
{
    OTF2_AttributeValue attribute_value;
    attribute_value.uint16 = uint16Value;

    /* Let OTF2_AttributeList_AddAttribute do the error checking */
    return OTF2_AttributeList_AddAttribute( attributeList,
                                            attributeID,
                                            OTF2_TYPE_UINT16,
                                            attribute_value );
}


OTF2_ErrorCode
OTF2_AttributeList_AddUint32( OTF2_AttributeList* attributeList,
                              OTF2_AttributeRef   attributeID,
                              uint32_t            uint32Value )
{
    OTF2_AttributeValue attribute_value;
    attribute_value.uint32 = uint32Value;

    /* Let OTF2_AttributeList_AddAttribute do the error checking */
    return OTF2_AttributeList_AddAttribute( attributeList,
                                            attributeID,
                                            OTF2_TYPE_UINT32,
                                            attribute_value );
}


OTF2_ErrorCode
OTF2_AttributeList_AddUint64( OTF2_AttributeList* attributeList,
                              OTF2_AttributeRef   attributeID,
                              uint64_t            uint64Value )
{
    OTF2_AttributeValue attribute_value;
    attribute_value.uint64 = uint64Value;

    /* Let OTF2_AttributeList_AddAttribute do the error checking */
    return OTF2_AttributeList_AddAttribute( attributeList,
                                            attributeID,
                                            OTF2_TYPE_UINT64,
                                            attribute_value );
}


OTF2_ErrorCode
OTF2_AttributeList_AddInt8( OTF2_AttributeList* attributeList,
                            OTF2_AttributeRef   attributeID,
                            int8_t              int8Value )
{
    OTF2_AttributeValue attribute_value;
    attribute_value.int8 = int8Value;

    /* Let OTF2_AttributeList_AddAttribute do the error checking */
    return OTF2_AttributeList_AddAttribute( attributeList,
                                            attributeID,
                                            OTF2_TYPE_INT8,
                                            attribute_value );
}


OTF2_ErrorCode
OTF2_AttributeList_AddInt16( OTF2_AttributeList* attributeList,
                             OTF2_AttributeRef   attributeID,
                             int16_t             int16Value )
{
    OTF2_AttributeValue attribute_value;
    attribute_value.int16 = int16Value;

    /* Let OTF2_AttributeList_AddAttribute do the error checking */
    return OTF2_AttributeList_AddAttribute( attributeList,
                                            attributeID,
                                            OTF2_TYPE_INT16,
                                            attribute_value );
}


OTF2_ErrorCode
OTF2_AttributeList_AddInt32( OTF2_AttributeList* attributeList,
                             OTF2_AttributeRef   attributeID,
                             int32_t             int32Value )
{
    OTF2_AttributeValue attribute_value;
    attribute_value.int32 = int32Value;

    /* Let OTF2_AttributeList_AddAttribute do the error checking */
    return OTF2_AttributeList_AddAttribute( attributeList,
                                            attributeID,
                                            OTF2_TYPE_INT32,
                                            attribute_value );
}


OTF2_ErrorCode
OTF2_AttributeList_AddInt64( OTF2_AttributeList* attributeList,
                             OTF2_AttributeRef   attributeID,
                             int64_t             int64Value )
{
    OTF2_AttributeValue attribute_value;
    attribute_value.int64 = int64Value;

    /* Let OTF2_AttributeList_AddAttribute do the error checking */
    return OTF2_AttributeList_AddAttribute( attributeList,
                                            attributeID,
                                            OTF2_TYPE_INT64,
                                            attribute_value );
}


OTF2_ErrorCode
OTF2_AttributeList_AddFloat( OTF2_AttributeList* attributeList,
                             OTF2_AttributeRef   attributeID,
                             float               float32Value )
{
    OTF2_AttributeValue attribute_value;
    attribute_value.float32 = float32Value;

    /* Let OTF2_AttributeList_AddAttribute do the error checking */
    return OTF2_AttributeList_AddAttribute( attributeList,
                                            attributeID,
                                            OTF2_TYPE_FLOAT,
                                            attribute_value );
}


OTF2_ErrorCode
OTF2_AttributeList_AddDouble( OTF2_AttributeList* attributeList,
                              OTF2_AttributeRef   attributeID,
                              double              float64Value )
{
    OTF2_AttributeValue attribute_value;
    attribute_value.float64 = float64Value;

    /* Let OTF2_AttributeList_AddAttribute do the error checking */
    return OTF2_AttributeList_AddAttribute( attributeList,
                                            attributeID,
                                            OTF2_TYPE_DOUBLE,
                                            attribute_value );
}


OTF2_ErrorCode
OTF2_AttributeList_AddStringRef( OTF2_AttributeList* attributeList,
                                 OTF2_AttributeRef   attributeID,
                                 OTF2_StringRef      stringRef )
{
    OTF2_AttributeValue attribute_value;
    attribute_value.stringRef = stringRef;

    /* Let OTF2_AttributeList_AddAttribute do the error checking */
    return OTF2_AttributeList_AddAttribute( attributeList,
                                            attributeID,
                                            OTF2_TYPE_STRING,
                                            attribute_value );
}


OTF2_ErrorCode
OTF2_AttributeList_AddAttributeRef( OTF2_AttributeList* attributeList,
                                    OTF2_AttributeRef   attributeID,
                                    OTF2_AttributeRef   attributeRef )
{
    OTF2_AttributeValue attribute_value;
    attribute_value.attributeRef = attributeRef;

    /* Let OTF2_AttributeList_AddAttribute do the error checking */
    return OTF2_AttributeList_AddAttribute( attributeList,
                                            attributeID,
                                            OTF2_TYPE_ATTRIBUTE,
                                            attribute_value );
}


OTF2_ErrorCode
OTF2_AttributeList_AddLocationRef( OTF2_AttributeList* attributeList,
                                   OTF2_AttributeRef   attributeID,
                                   OTF2_LocationRef    locationRef )
{
    OTF2_AttributeValue attribute_value;
    attribute_value.locationRef = locationRef;

    /* Let OTF2_AttributeList_AddAttribute do the error checking */
    return OTF2_AttributeList_AddAttribute( attributeList,
                                            attributeID,
                                            OTF2_TYPE_LOCATION,
                                            attribute_value );
}


OTF2_ErrorCode
OTF2_AttributeList_AddRegionRef( OTF2_AttributeList* attributeList,
                                 OTF2_AttributeRef   attributeID,
                                 OTF2_RegionRef      regionRef )
{
    OTF2_AttributeValue attribute_value;
    attribute_value.regionRef = regionRef;

    /* Let OTF2_AttributeList_AddAttribute do the error checking */
    return OTF2_AttributeList_AddAttribute( attributeList,
                                            attributeID,
                                            OTF2_TYPE_REGION,
                                            attribute_value );
}


OTF2_ErrorCode
OTF2_AttributeList_AddGroupRef( OTF2_AttributeList* attributeList,
                                OTF2_AttributeRef   attributeID,
                                OTF2_GroupRef       groupRef )
{
    OTF2_AttributeValue attribute_value;
    attribute_value.groupRef = groupRef;

    /* Let OTF2_AttributeList_AddAttribute do the error checking */
    return OTF2_AttributeList_AddAttribute( attributeList,
                                            attributeID,
                                            OTF2_TYPE_GROUP,
                                            attribute_value );
}


OTF2_ErrorCode
OTF2_AttributeList_AddMetricRef( OTF2_AttributeList* attributeList,
                                 OTF2_AttributeRef   attributeID,
                                 OTF2_MetricRef      metricRef )
{
    OTF2_AttributeValue attribute_value;
    attribute_value.metricRef = metricRef;

    /* Let OTF2_AttributeList_AddAttribute do the error checking */
    return OTF2_AttributeList_AddAttribute( attributeList,
                                            attributeID,
                                            OTF2_TYPE_METRIC,
                                            attribute_value );
}


OTF2_ErrorCode
OTF2_AttributeList_AddCommRef( OTF2_AttributeList* attributeList,
                               OTF2_AttributeRef   attributeID,
                               OTF2_CommRef        commRef )
{
    OTF2_AttributeValue attribute_value;
    attribute_value.commRef = commRef;

    /* Let OTF2_AttributeList_AddAttribute do the error checking */
    return OTF2_AttributeList_AddAttribute( attributeList,
                                            attributeID,
                                            OTF2_TYPE_COMM,
                                            attribute_value );
}


OTF2_ErrorCode
OTF2_AttributeList_AddParameterRef( OTF2_AttributeList* attributeList,
                                    OTF2_AttributeRef   attributeID,
                                    OTF2_ParameterRef   parameterRef )
{
    OTF2_AttributeValue attribute_value;
    attribute_value.parameterRef = parameterRef;

    /* Let OTF2_AttributeList_AddAttribute do the error checking */
    return OTF2_AttributeList_AddAttribute( attributeList,
                                            attributeID,
                                            OTF2_TYPE_PARAMETER,
                                            attribute_value );
}


OTF2_ErrorCode
OTF2_AttributeList_AddRmaWinRef( OTF2_AttributeList* attributeList,
                                 OTF2_AttributeRef   attributeID,
                                 OTF2_RmaWinRef      rmaWinRef )
{
    OTF2_AttributeValue attribute_value;
    attribute_value.rmaWinRef = rmaWinRef;

    /* Let OTF2_AttributeList_AddAttribute do the error checking */
    return OTF2_AttributeList_AddAttribute( attributeList,
                                            attributeID,
                                            OTF2_TYPE_RMA_WIN,
                                            attribute_value );
}


OTF2_ErrorCode
OTF2_AttributeList_GetUint8( const OTF2_AttributeList* attributeList,
                             OTF2_AttributeRef         attributeID,
                             uint8_t*                  uint8Value )
{
    if ( !uint8Value )
    {
        return UTILS_ERROR( OTF2_ERROR_INVALID_ARGUMENT,
                            "NULL pointer for uint8Value." );
    }

    OTF2_Type           type;
    OTF2_AttributeValue attribute_value;

    OTF2_ErrorCode status =
        OTF2_AttributeList_GetAttributeByID( attributeList,
                                             attributeID,
                                             &type,
                                             &attribute_value );

    if ( status != OTF2_SUCCESS )
    {
        return status;
    }

    if ( type != OTF2_TYPE_UINT8 )
    {
        return UTILS_ERROR( OTF2_ERROR_INVALID_DATA,
                            "Requested value does not match type." );
    }

    *uint8Value = attribute_value.uint8;

    return OTF2_SUCCESS;
}


OTF2_ErrorCode
OTF2_AttributeList_GetUint16( const OTF2_AttributeList* attributeList,
                              OTF2_AttributeRef         attributeID,
                              uint16_t*                 uint16Value )
{
    if ( !uint16Value )
    {
        return UTILS_ERROR( OTF2_ERROR_INVALID_ARGUMENT,
                            "NULL pointer for uint16Value." );
    }

    OTF2_Type           type;
    OTF2_AttributeValue attribute_value;

    OTF2_ErrorCode status =
        OTF2_AttributeList_GetAttributeByID( attributeList,
                                             attributeID,
                                             &type,
                                             &attribute_value );

    if ( status != OTF2_SUCCESS )
    {
        return status;
    }

    if ( type != OTF2_TYPE_UINT16 )
    {
        return UTILS_ERROR( OTF2_ERROR_INVALID_DATA,
                            "Requested value does not match type." );
    }

    *uint16Value = attribute_value.uint16;

    return OTF2_SUCCESS;
}


OTF2_ErrorCode
OTF2_AttributeList_GetUint32( const OTF2_AttributeList* attributeList,
                              OTF2_AttributeRef         attributeID,
                              uint32_t*                 uint32Value )
{
    if ( !uint32Value )
    {
        return UTILS_ERROR( OTF2_ERROR_INVALID_ARGUMENT,
                            "NULL pointer for uint32Value." );
    }

    OTF2_Type           type;
    OTF2_AttributeValue attribute_value;

    OTF2_ErrorCode status =
        OTF2_AttributeList_GetAttributeByID( attributeList,
                                             attributeID,
                                             &type,
                                             &attribute_value );

    if ( status != OTF2_SUCCESS )
    {
        return status;
    }

    if ( type != OTF2_TYPE_UINT32 )
    {
        return UTILS_ERROR( OTF2_ERROR_INVALID_DATA,
                            "Requested value does not match type." );
    }

    *uint32Value = attribute_value.uint32;

    return OTF2_SUCCESS;
}


OTF2_ErrorCode
OTF2_AttributeList_GetUint64( const OTF2_AttributeList* attributeList,
                              OTF2_AttributeRef         attributeID,
                              uint64_t*                 uint64Value )
{
    if ( !uint64Value )
    {
        return UTILS_ERROR( OTF2_ERROR_INVALID_ARGUMENT,
                            "NULL pointer for uint64Value." );
    }

    OTF2_Type           type;
    OTF2_AttributeValue attribute_value;

    OTF2_ErrorCode status =
        OTF2_AttributeList_GetAttributeByID( attributeList,
                                             attributeID,
                                             &type,
                                             &attribute_value );

    if ( status != OTF2_SUCCESS )
    {
        return status;
    }

    if ( type != OTF2_TYPE_UINT64 )
    {
        return UTILS_ERROR( OTF2_ERROR_INVALID_DATA,
                            "Requested value does not match type." );
    }

    *uint64Value = attribute_value.uint64;

    return OTF2_SUCCESS;
}


OTF2_ErrorCode
OTF2_AttributeList_GetInt8( const OTF2_AttributeList* attributeList,
                            OTF2_AttributeRef         attributeID,
                            int8_t*                   int8Value )
{
    if ( !int8Value )
    {
        return UTILS_ERROR( OTF2_ERROR_INVALID_ARGUMENT,
                            "NULL pointer for int8Value." );
    }

    OTF2_Type           type;
    OTF2_AttributeValue attribute_value;

    OTF2_ErrorCode status =
        OTF2_AttributeList_GetAttributeByID( attributeList,
                                             attributeID,
                                             &type,
                                             &attribute_value );

    if ( status != OTF2_SUCCESS )
    {
        return status;
    }

    if ( type != OTF2_TYPE_INT8 )
    {
        return UTILS_ERROR( OTF2_ERROR_INVALID_DATA,
                            "Requested value does not match type." );
    }

    *int8Value = attribute_value.int8;

    return OTF2_SUCCESS;
}


OTF2_ErrorCode
OTF2_AttributeList_GetInt16( const OTF2_AttributeList* attributeList,
                             OTF2_AttributeRef         attributeID,
                             int16_t*                  int16Value )
{
    if ( !int16Value )
    {
        return UTILS_ERROR( OTF2_ERROR_INVALID_ARGUMENT,
                            "NULL pointer for int16Value." );
    }

    OTF2_Type           type;
    OTF2_AttributeValue attribute_value;

    OTF2_ErrorCode status =
        OTF2_AttributeList_GetAttributeByID( attributeList,
                                             attributeID,
                                             &type,
                                             &attribute_value );

    if ( status != OTF2_SUCCESS )
    {
        return status;
    }

    if ( type != OTF2_TYPE_INT16 )
    {
        return UTILS_ERROR( OTF2_ERROR_INVALID_DATA,
                            "Requested value does not match type." );
    }

    *int16Value = attribute_value.int16;

    return OTF2_SUCCESS;
}


OTF2_ErrorCode
OTF2_AttributeList_GetInt32( const OTF2_AttributeList* attributeList,
                             OTF2_AttributeRef         attributeID,
                             int32_t*                  int32Value )
{
    if ( !int32Value )
    {
        return UTILS_ERROR( OTF2_ERROR_INVALID_ARGUMENT,
                            "NULL pointer for int32Value." );
    }

    OTF2_Type           type;
    OTF2_AttributeValue attribute_value;

    OTF2_ErrorCode status =
        OTF2_AttributeList_GetAttributeByID( attributeList,
                                             attributeID,
                                             &type,
                                             &attribute_value );

    if ( status != OTF2_SUCCESS )
    {
        return status;
    }

    if ( type != OTF2_TYPE_INT32 )
    {
        return UTILS_ERROR( OTF2_ERROR_INVALID_DATA,
                            "Requested value does not match type." );
    }

    *int32Value = attribute_value.int32;

    return OTF2_SUCCESS;
}


OTF2_ErrorCode
OTF2_AttributeList_GetInt64( const OTF2_AttributeList* attributeList,
                             OTF2_AttributeRef         attributeID,
                             int64_t*                  int64Value )
{
    if ( !int64Value )
    {
        return UTILS_ERROR( OTF2_ERROR_INVALID_ARGUMENT,
                            "NULL pointer for int64Value." );
    }

    OTF2_Type           type;
    OTF2_AttributeValue attribute_value;

    OTF2_ErrorCode status =
        OTF2_AttributeList_GetAttributeByID( attributeList,
                                             attributeID,
                                             &type,
                                             &attribute_value );

    if ( status != OTF2_SUCCESS )
    {
        return status;
    }

    if ( type != OTF2_TYPE_INT64 )
    {
        return UTILS_ERROR( OTF2_ERROR_INVALID_DATA,
                            "Requested value does not match type." );
    }

    *int64Value = attribute_value.int64;

    return OTF2_SUCCESS;
}


OTF2_ErrorCode
OTF2_AttributeList_GetFloat( const OTF2_AttributeList* attributeList,
                             OTF2_AttributeRef         attributeID,
                             float*                    float32Value )
{
    if ( !float32Value )
    {
        return UTILS_ERROR( OTF2_ERROR_INVALID_ARGUMENT,
                            "NULL pointer for float32Value." );
    }

    OTF2_Type           type;
    OTF2_AttributeValue attribute_value;

    OTF2_ErrorCode status =
        OTF2_AttributeList_GetAttributeByID( attributeList,
                                             attributeID,
                                             &type,
                                             &attribute_value );

    if ( status != OTF2_SUCCESS )
    {
        return status;
    }

    if ( type != OTF2_TYPE_FLOAT )
    {
        return UTILS_ERROR( OTF2_ERROR_INVALID_DATA,
                            "Requested value does not match type." );
    }

    *float32Value = attribute_value.float32;

    return OTF2_SUCCESS;
}


OTF2_ErrorCode
OTF2_AttributeList_GetDouble( const OTF2_AttributeList* attributeList,
                              OTF2_AttributeRef         attributeID,
                              double*                   float64Value )
{
    if ( !float64Value )
    {
        return UTILS_ERROR( OTF2_ERROR_INVALID_ARGUMENT,
                            "NULL pointer for float64Value." );
    }

    OTF2_Type           type;
    OTF2_AttributeValue attribute_value;

    OTF2_ErrorCode status =
        OTF2_AttributeList_GetAttributeByID( attributeList,
                                             attributeID,
                                             &type,
                                             &attribute_value );

    if ( status != OTF2_SUCCESS )
    {
        return status;
    }

    if ( type != OTF2_TYPE_DOUBLE )
    {
        return UTILS_ERROR( OTF2_ERROR_INVALID_DATA,
                            "Requested value does not match type." );
    }

    *float64Value = attribute_value.float64;

    return OTF2_SUCCESS;
}


OTF2_ErrorCode
OTF2_AttributeList_GetStringRef( const OTF2_AttributeList* attributeList,
                                 OTF2_AttributeRef         attributeID,
                                 OTF2_StringRef*           stringRef )
{
    if ( !stringRef )
    {
        return UTILS_ERROR( OTF2_ERROR_INVALID_ARGUMENT,
                            "NULL pointer for stringRef." );
    }

    OTF2_Type           type;
    OTF2_AttributeValue attribute_value;

    OTF2_ErrorCode status =
        OTF2_AttributeList_GetAttributeByID( attributeList,
                                             attributeID,
                                             &type,
                                             &attribute_value );

    if ( status != OTF2_SUCCESS )
    {
        return status;
    }

    if ( type != OTF2_TYPE_STRING )
    {
        return UTILS_ERROR( OTF2_ERROR_INVALID_DATA,
                            "Requested value does not match type." );
    }

    *stringRef = attribute_value.stringRef;

    return OTF2_SUCCESS;
}


OTF2_ErrorCode
OTF2_AttributeList_GetAttributeRef( const OTF2_AttributeList* attributeList,
                                    OTF2_AttributeRef         attributeID,
                                    OTF2_AttributeRef*        attributeRef )
{
    if ( !attributeRef )
    {
        return UTILS_ERROR( OTF2_ERROR_INVALID_ARGUMENT,
                            "NULL pointer for attributeRef." );
    }

    OTF2_Type           type;
    OTF2_AttributeValue attribute_value;

    OTF2_ErrorCode status =
        OTF2_AttributeList_GetAttributeByID( attributeList,
                                             attributeID,
                                             &type,
                                             &attribute_value );

    if ( status != OTF2_SUCCESS )
    {
        return status;
    }

    if ( type != OTF2_TYPE_ATTRIBUTE )
    {
        return UTILS_ERROR( OTF2_ERROR_INVALID_DATA,
                            "Requested value does not match type." );
    }

    *attributeRef = attribute_value.attributeRef;

    return OTF2_SUCCESS;
}


OTF2_ErrorCode
OTF2_AttributeList_GetLocationRef( const OTF2_AttributeList* attributeList,
                                   OTF2_AttributeRef         attributeID,
                                   OTF2_LocationRef*         locationRef )
{
    if ( !locationRef )
    {
        return UTILS_ERROR( OTF2_ERROR_INVALID_ARGUMENT,
                            "NULL pointer for locationRef." );
    }

    OTF2_Type           type;
    OTF2_AttributeValue attribute_value;

    OTF2_ErrorCode status =
        OTF2_AttributeList_GetAttributeByID( attributeList,
                                             attributeID,
                                             &type,
                                             &attribute_value );

    if ( status != OTF2_SUCCESS )
    {
        return status;
    }

    if ( type != OTF2_TYPE_LOCATION )
    {
        return UTILS_ERROR( OTF2_ERROR_INVALID_DATA,
                            "Requested value does not match type." );
    }

    *locationRef = attribute_value.locationRef;

    return OTF2_SUCCESS;
}


OTF2_ErrorCode
OTF2_AttributeList_GetRegionRef( const OTF2_AttributeList* attributeList,
                                 OTF2_AttributeRef         attributeID,
                                 OTF2_RegionRef*           regionRef )
{
    if ( !regionRef )
    {
        return UTILS_ERROR( OTF2_ERROR_INVALID_ARGUMENT,
                            "NULL pointer for regionRef." );
    }

    OTF2_Type           type;
    OTF2_AttributeValue attribute_value;

    OTF2_ErrorCode status =
        OTF2_AttributeList_GetAttributeByID( attributeList,
                                             attributeID,
                                             &type,
                                             &attribute_value );

    if ( status != OTF2_SUCCESS )
    {
        return status;
    }

    if ( type != OTF2_TYPE_REGION )
    {
        return UTILS_ERROR( OTF2_ERROR_INVALID_DATA,
                            "Requested value does not match type." );
    }

    *regionRef = attribute_value.regionRef;

    return OTF2_SUCCESS;
}


OTF2_ErrorCode
OTF2_AttributeList_GetGroupRef( const OTF2_AttributeList* attributeList,
                                OTF2_AttributeRef         attributeID,
                                OTF2_GroupRef*            groupRef )
{
    if ( !groupRef )
    {
        return UTILS_ERROR( OTF2_ERROR_INVALID_ARGUMENT,
                            "NULL pointer for groupRef." );
    }

    OTF2_Type           type;
    OTF2_AttributeValue attribute_value;

    OTF2_ErrorCode status =
        OTF2_AttributeList_GetAttributeByID( attributeList,
                                             attributeID,
                                             &type,
                                             &attribute_value );

    if ( status != OTF2_SUCCESS )
    {
        return status;
    }

    if ( type != OTF2_TYPE_GROUP )
    {
        return UTILS_ERROR( OTF2_ERROR_INVALID_DATA,
                            "Requested value does not match type." );
    }

    *groupRef = attribute_value.groupRef;

    return OTF2_SUCCESS;
}


OTF2_ErrorCode
OTF2_AttributeList_GetMetricRef( const OTF2_AttributeList* attributeList,
                                 OTF2_AttributeRef         attributeID,
                                 OTF2_MetricRef*           metricRef )
{
    if ( !metricRef )
    {
        return UTILS_ERROR( OTF2_ERROR_INVALID_ARGUMENT,
                            "NULL pointer for metricRef." );
    }

    OTF2_Type           type;
    OTF2_AttributeValue attribute_value;

    OTF2_ErrorCode status =
        OTF2_AttributeList_GetAttributeByID( attributeList,
                                             attributeID,
                                             &type,
                                             &attribute_value );

    if ( status != OTF2_SUCCESS )
    {
        return status;
    }

    if ( type != OTF2_TYPE_METRIC )
    {
        return UTILS_ERROR( OTF2_ERROR_INVALID_DATA,
                            "Requested value does not match type." );
    }

    *metricRef = attribute_value.metricRef;

    return OTF2_SUCCESS;
}


OTF2_ErrorCode
OTF2_AttributeList_GetCommRef( const OTF2_AttributeList* attributeList,
                               OTF2_AttributeRef         attributeID,
                               OTF2_CommRef*             commRef )
{
    if ( !commRef )
    {
        return UTILS_ERROR( OTF2_ERROR_INVALID_ARGUMENT,
                            "NULL pointer for commRef." );
    }

    OTF2_Type           type;
    OTF2_AttributeValue attribute_value;

    OTF2_ErrorCode status =
        OTF2_AttributeList_GetAttributeByID( attributeList,
                                             attributeID,
                                             &type,
                                             &attribute_value );

    if ( status != OTF2_SUCCESS )
    {
        return status;
    }

    if ( type != OTF2_TYPE_COMM )
    {
        return UTILS_ERROR( OTF2_ERROR_INVALID_DATA,
                            "Requested value does not match type." );
    }

    *commRef = attribute_value.commRef;

    return OTF2_SUCCESS;
}


OTF2_ErrorCode
OTF2_AttributeList_GetParameterRef( const OTF2_AttributeList* attributeList,
                                    OTF2_AttributeRef         attributeID,
                                    OTF2_ParameterRef*        parameterRef )
{
    if ( !parameterRef )
    {
        return UTILS_ERROR( OTF2_ERROR_INVALID_ARGUMENT,
                            "NULL pointer for parameterRef." );
    }

    OTF2_Type           type;
    OTF2_AttributeValue attribute_value;

    OTF2_ErrorCode status =
        OTF2_AttributeList_GetAttributeByID( attributeList,
                                             attributeID,
                                             &type,
                                             &attribute_value );

    if ( status != OTF2_SUCCESS )
    {
        return status;
    }

    if ( type != OTF2_TYPE_PARAMETER )
    {
        return UTILS_ERROR( OTF2_ERROR_INVALID_DATA,
                            "Requested value does not match type." );
    }

    *parameterRef = attribute_value.parameterRef;

    return OTF2_SUCCESS;
}


OTF2_ErrorCode
OTF2_AttributeList_GetRmaWinRef( const OTF2_AttributeList* attributeList,
                                 OTF2_AttributeRef         attributeID,
                                 OTF2_RmaWinRef*           rmaWinRef )
{
    if ( !rmaWinRef )
    {
        return UTILS_ERROR( OTF2_ERROR_INVALID_ARGUMENT,
                            "NULL pointer for rmaWinRef." );
    }

    OTF2_Type           type;
    OTF2_AttributeValue attribute_value;

    OTF2_ErrorCode status =
        OTF2_AttributeList_GetAttributeByID( attributeList,
                                             attributeID,
                                             &type,
                                             &attribute_value );

    if ( status != OTF2_SUCCESS )
    {
        return status;
    }

    if ( type != OTF2_TYPE_RMA_WIN )
    {
        return UTILS_ERROR( OTF2_ERROR_INVALID_DATA,
                            "Requested value does not match type." );
    }

    *rmaWinRef = attribute_value.rmaWinRef;

    return OTF2_SUCCESS;
}


OTF2_ErrorCode
otf2_attribute_list_read_from_buffer( OTF2_AttributeList*  attributeList,
                                      OTF2_Buffer*         buffer,
                                      OTF2_MappingCallback mappingCb,
                                      void*                mappingData )
{
    UTILS_ASSERT( attributeList );
    UTILS_ASSERT( buffer );

    OTF2_ErrorCode status;

    /* Get record length and test memory availability */
    uint64_t record_data_length;
    status = OTF2_Buffer_GuaranteeRecord( buffer, &record_data_length );
    if ( OTF2_SUCCESS != status )
    {
        return UTILS_ERROR( status,
                            "Could not read attribute list. "
                            "Not enough memory in buffer." );
    }
    uint8_t* record_start_pos;
    OTF2_Buffer_GetPosition( buffer, &record_start_pos );

    /* Get how many attributes are in the list */
    uint32_t number_of_attributes;
    status = OTF2_Buffer_ReadUint32( buffer, &number_of_attributes );
    if ( OTF2_SUCCESS != status )
    {
        return UTILS_ERROR( status,
                            "Could not read number of attributes in list. "
                            "Invalid compression size." );
    }

    /* Check if number of attributes is consistent with record length */
    if ( record_data_length > number_of_attributes * OTF2_MAXIMUM_ATTRIBUTE_LIST_ENTRY_SIZE + 5 )
    {
        return UTILS_ERROR( OTF2_ERROR_INTEGRITY_FAULT,
                            "Could not read attribute list. "
                            "Record length and number of attributes in list are not consistent." );
    }

    for ( uint32_t i = 0; i < number_of_attributes; i++ )
    {
        uint32_t            attribute_id;
        OTF2_Type           type;
        OTF2_AttributeValue value;

        /* Read attribute id from trace */
        status = OTF2_Buffer_ReadUint32( buffer, &attribute_id );
        if ( OTF2_SUCCESS != status )
        {
            return UTILS_ERROR( status,
                                "Could not read attribute from list. "
                                "Invalid compression size." );
        }
        attribute_id = mappingCb( mappingData, OTF2_MAPPING_ATTRIBUTE, attribute_id );

        /* Read the type from trace */
        OTF2_Buffer_ReadUint8( buffer, &type );

        /* Cast the container to the right type and read it from the trace */
        switch ( type )
        {
            case OTF2_TYPE_UINT8:
                OTF2_Buffer_ReadUint8( buffer, &value.uint8 );
                break;

            case OTF2_TYPE_UINT16:
                OTF2_Buffer_ReadUint16( buffer, &value.uint16 );
                break;

            case OTF2_TYPE_UINT32:
                status = OTF2_Buffer_ReadUint32( buffer, &value.uint32 );
                if ( OTF2_SUCCESS != status )
                {
                    return UTILS_ERROR( status,
                                        "Could not read attribute. "
                                        "Invalid compression size." );
                }
                break;

            case OTF2_TYPE_UINT64:
                status = OTF2_Buffer_ReadUint64( buffer, &value.uint64 );
                if ( OTF2_SUCCESS != status )
                {
                    return UTILS_ERROR( status,
                                        "Could not read attribute. "
                                        "Invalid compression size." );
                }
                break;

            case OTF2_TYPE_INT8:
                OTF2_Buffer_ReadInt8( buffer, &value.int8 );
                break;

            case OTF2_TYPE_INT16:
                OTF2_Buffer_ReadInt16( buffer, &value.int16 );
                break;

            case OTF2_TYPE_INT32:
                status = OTF2_Buffer_ReadInt32( buffer, &value.int32 );
                if ( OTF2_SUCCESS != status )
                {
                    return UTILS_ERROR( status,
                                        "Could not read attribute. "
                                        "Invalid compression size." );
                }
                break;

            case OTF2_TYPE_INT64:
                status = OTF2_Buffer_ReadInt64( buffer, &value.int64 );
                if ( OTF2_SUCCESS != status )
                {
                    return UTILS_ERROR( status,
                                        "Could not read attribute. "
                                        "Invalid compression size." );
                }
                break;

            case OTF2_TYPE_FLOAT:
                OTF2_Buffer_ReadFloat( buffer, &value.float32 );
                break;

            case OTF2_TYPE_DOUBLE:
                OTF2_Buffer_ReadDouble( buffer, &value.float64 );
                break;

            case OTF2_TYPE_STRING:
                /* OTF2_TYPE_STRING attributes reference a String definition */
                status = OTF2_Buffer_ReadUint32( buffer,
                                                 &value.stringRef );
                if ( OTF2_SUCCESS != status )
                {
                    return UTILS_ERROR( status,
                                        "Could not read String attribute. "
                                        "Invalid compression size." );
                }
                /* But we need to map them to the global definition */
                value.stringRef = mappingCb( mappingData,
                                             OTF2_MAPPING_STRING,
                                             value.stringRef );
                break;

            case OTF2_TYPE_ATTRIBUTE:
                /* OTF2_TYPE_ATTRIBUTE attributes reference a Attribute definition */
                status = OTF2_Buffer_ReadUint32( buffer,
                                                 &value.attributeRef );
                if ( OTF2_SUCCESS != status )
                {
                    return UTILS_ERROR( status,
                                        "Could not read Attribute attribute. "
                                        "Invalid compression size." );
                }
                /* But we need to map them to the global definition */
                value.attributeRef = mappingCb( mappingData,
                                                OTF2_MAPPING_ATTRIBUTE,
                                                value.attributeRef );
                break;

            case OTF2_TYPE_LOCATION:
                /* OTF2_TYPE_LOCATION attributes reference a Location definition */
                status = OTF2_Buffer_ReadUint64( buffer,
                                                 &value.locationRef );
                if ( OTF2_SUCCESS != status )
                {
                    return UTILS_ERROR( status,
                                        "Could not read Location attribute. "
                                        "Invalid compression size." );
                }
                /* But we need to map them to the global definition */
                value.locationRef = mappingCb( mappingData,
                                               OTF2_MAPPING_LOCATION,
                                               value.locationRef );
                break;

            case OTF2_TYPE_REGION:
                /* OTF2_TYPE_REGION attributes reference a Region definition */
                status = OTF2_Buffer_ReadUint32( buffer,
                                                 &value.regionRef );
                if ( OTF2_SUCCESS != status )
                {
                    return UTILS_ERROR( status,
                                        "Could not read Region attribute. "
                                        "Invalid compression size." );
                }
                /* But we need to map them to the global definition */
                value.regionRef = mappingCb( mappingData,
                                             OTF2_MAPPING_REGION,
                                             value.regionRef );
                break;

            case OTF2_TYPE_GROUP:
                /* OTF2_TYPE_GROUP attributes reference a Group definition */
                status = OTF2_Buffer_ReadUint32( buffer,
                                                 &value.groupRef );
                if ( OTF2_SUCCESS != status )
                {
                    return UTILS_ERROR( status,
                                        "Could not read Group attribute. "
                                        "Invalid compression size." );
                }
                /* But we need to map them to the global definition */
                value.groupRef = mappingCb( mappingData,
                                            OTF2_MAPPING_GROUP,
                                            value.groupRef );
                break;

            case OTF2_TYPE_METRIC:
                /* OTF2_TYPE_METRIC attributes reference a MetricClass, or a MetricInstance definition */
                status = OTF2_Buffer_ReadUint32( buffer,
                                                 &value.metricRef );
                if ( OTF2_SUCCESS != status )
                {
                    return UTILS_ERROR( status,
                                        "Could not read MetricClass attribute. "
                                        "Invalid compression size." );
                }
                /* But we need to map them to the global definition */
                value.metricRef = mappingCb( mappingData,
                                             OTF2_MAPPING_METRIC,
                                             value.metricRef );
                break;

            case OTF2_TYPE_COMM:
                /* OTF2_TYPE_COMM attributes reference a Comm definition */
                status = OTF2_Buffer_ReadUint32( buffer,
                                                 &value.commRef );
                if ( OTF2_SUCCESS != status )
                {
                    return UTILS_ERROR( status,
                                        "Could not read Comm attribute. "
                                        "Invalid compression size." );
                }
                /* But we need to map them to the global definition */
                value.commRef = mappingCb( mappingData,
                                           OTF2_MAPPING_COMM,
                                           value.commRef );
                break;

            case OTF2_TYPE_PARAMETER:
                /* OTF2_TYPE_PARAMETER attributes reference a Parameter definition */
                status = OTF2_Buffer_ReadUint32( buffer,
                                                 &value.parameterRef );
                if ( OTF2_SUCCESS != status )
                {
                    return UTILS_ERROR( status,
                                        "Could not read Parameter attribute. "
                                        "Invalid compression size." );
                }
                /* But we need to map them to the global definition */
                value.parameterRef = mappingCb( mappingData,
                                                OTF2_MAPPING_PARAMETER,
                                                value.parameterRef );
                break;

            case OTF2_TYPE_RMA_WIN:
                /* OTF2_TYPE_RMA_WIN attributes reference a RmaWin definition */
                status = OTF2_Buffer_ReadUint32( buffer,
                                                 &value.rmaWinRef );
                if ( OTF2_SUCCESS != status )
                {
                    return UTILS_ERROR( status,
                                        "Could not read RmaWin attribute. "
                                        "Invalid compression size." );
                }
                /* But we need to map them to the global definition */
                value.rmaWinRef = mappingCb( mappingData,
                                             OTF2_MAPPING_RMA_WIN,
                                             value.rmaWinRef );
                break;

            default:
                /*
                 * There can only be more definition references added in the future,
                 * and definition references are guaranteed to be of compressable
                 * size, thus we induce that the value is of compressed type
                 */
                status = OTF2_Buffer_ReadUint64( buffer, &value.uint64 );
                if ( OTF2_SUCCESS != status )
                {
                    return UTILS_ERROR( status,
                                        "Could not read attribute of unknown type. "
                                        "Invalid compression size." );
                }
                break;
        }

        /* Store the new value in the attribute list. */
        OTF2_AttributeList_AddAttribute( attributeList,
                                         attribute_id,
                                         type,
                                         value );
    }

    /*
     * Always jump to the announced end of the record.
     */
    status = OTF2_Buffer_SetPosition( buffer, record_start_pos + record_data_length );
    if ( OTF2_SUCCESS != status )
    {
        return UTILS_ERROR( status, "Could not read attribute list." );
    }

    return OTF2_SUCCESS;
}


OTF2_ErrorCode
otf2_attribute_list_write_to_buffer( OTF2_AttributeList* attributeList,
                                     OTF2_Buffer*        buffer )
{
    /* Write no attribute list to the buffer, if there is no one. */
    if ( attributeList == NULL || attributeList->capacity == 0 )
    {
        return OTF2_SUCCESS;
    }

    /* Write the attribute lists record type id to the trace */
    OTF2_Buffer_WriteUint8( buffer, OTF2_ATTRIBUTE_LIST );

    /* Write initial record length */
    uint32_t data_size_of_attr_list = otf2_attribute_list_get_data_size( attributeList );
    OTF2_Buffer_WriteInitialRecordLength( buffer, data_size_of_attr_list );

    /* Store how many items are in the list to the trace */
    OTF2_Buffer_WriteUint32( buffer, attributeList->capacity );

    for ( otf2_attribute* attr = attributeList->head; attr; attr = attr->next )
    {
        /* Write attribute id to the trace */
        OTF2_Buffer_WriteUint32( buffer, attr->attribute_id );

        /* Write the type to the trace */
        OTF2_Buffer_WriteUint8( buffer, attr->type_id );

        switch ( attr->type_id )
        {
            case OTF2_TYPE_UINT8:
                OTF2_Buffer_WriteUint8( buffer,
                                        attr->value.uint8 );
                break;

            case OTF2_TYPE_UINT16:
                OTF2_Buffer_WriteUint16( buffer,
                                         attr->value.uint16 );
                break;

            case OTF2_TYPE_UINT32:
                OTF2_Buffer_WriteUint32( buffer,
                                         attr->value.uint32 );
                break;

            case OTF2_TYPE_UINT64:
                OTF2_Buffer_WriteUint64( buffer,
                                         attr->value.uint64 );
                break;

            case OTF2_TYPE_INT8:
                OTF2_Buffer_WriteInt8( buffer,
                                       attr->value.int8 );
                break;

            case OTF2_TYPE_INT16:
                OTF2_Buffer_WriteInt16( buffer,
                                        attr->value.int16 );
                break;

            case OTF2_TYPE_INT32:
                OTF2_Buffer_WriteInt32( buffer,
                                        attr->value.int32 );
                break;

            case OTF2_TYPE_INT64:
                OTF2_Buffer_WriteInt64( buffer,
                                        attr->value.int64 );
                break;

            case OTF2_TYPE_FLOAT:
                OTF2_Buffer_WriteFloat( buffer,
                                        attr->value.float32 );
                break;

            case OTF2_TYPE_DOUBLE:
                OTF2_Buffer_WriteDouble( buffer,
                                         attr->value.float64 );
                break;

            case OTF2_TYPE_STRING:
                /* OTF2_TYPE_STRING attributes reference a String definition */
                OTF2_Buffer_WriteUint32( buffer, attr->value.stringRef );
                break;

            case OTF2_TYPE_ATTRIBUTE:
                /* OTF2_TYPE_ATTRIBUTE attributes reference a Attribute definition */
                OTF2_Buffer_WriteUint32( buffer, attr->value.attributeRef );
                break;

            case OTF2_TYPE_LOCATION:
                /* OTF2_TYPE_LOCATION attributes reference a Location definition */
                OTF2_Buffer_WriteUint64( buffer, attr->value.locationRef );
                break;

            case OTF2_TYPE_REGION:
                /* OTF2_TYPE_REGION attributes reference a Region definition */
                OTF2_Buffer_WriteUint32( buffer, attr->value.regionRef );
                break;

            case OTF2_TYPE_GROUP:
                /* OTF2_TYPE_GROUP attributes reference a Group definition */
                OTF2_Buffer_WriteUint32( buffer, attr->value.groupRef );
                break;

            case OTF2_TYPE_METRIC:
                /* OTF2_TYPE_METRIC attributes reference a MetricClass, or a MetricInstance definition */
                OTF2_Buffer_WriteUint32( buffer, attr->value.metricRef );
                break;

            case OTF2_TYPE_COMM:
                /* OTF2_TYPE_COMM attributes reference a Comm definition */
                OTF2_Buffer_WriteUint32( buffer, attr->value.commRef );
                break;

            case OTF2_TYPE_PARAMETER:
                /* OTF2_TYPE_PARAMETER attributes reference a Parameter definition */
                OTF2_Buffer_WriteUint32( buffer, attr->value.parameterRef );
                break;

            case OTF2_TYPE_RMA_WIN:
                /* OTF2_TYPE_RMA_WIN attributes reference a RmaWin definition */
                OTF2_Buffer_WriteUint32( buffer, attr->value.rmaWinRef );
                break;

            default:
                /*
                 * When getting an unknown type, we write it nevertheless as an
                 * uint64, because we assume that future additions will always be
                 * definition references and they are always of compressed types.
                 */
                OTF2_Buffer_WriteUint64( buffer, attr->value.uint64 );
                break;
        }
    }

    otf2_attribute_list_remove_all_attributes( attributeList );

    /* Write real record length */
    return OTF2_Buffer_WriteFinalRecordLength( buffer, data_size_of_attr_list );
}
